--Copyright 2016 Josiah Goeman

print("Physgun Info Monitor V2.0")

local HOMEBREW_MODEL_PATH = "models/homebrew/homebrew.mdl"
local CRT_MODEL_PATH = "models/crt/crt.mdl"
local HOLO_MODEL_PATH = "models/holo/holo.mdl"

if SERVER then
	--make clients download model files
	resource.AddFile("materials/physgun_info_monitor/context_icon.png")
	
	resource.AddFile("materials/models/homebrew/frame_mat.vmt")
	resource.AddFile("materials/models/homebrew/led_mat.vmt")
	resource.AddFile("materials/models/homebrew/ribbon_mat.vmt")
	resource.AddFile("materials/models/homebrew/screen_mat.vmt")
	resource.AddFile("materials/models/homebrew/homebrew_diffuse.vtf")
	resource.AddFile("materials/models/homebrew/homebrew_normal.vtf")
	resource.AddFile("models/homebrew/homebrew.mdl")
	
	resource.AddFile("materials/models/crt/frame_mat.vmt")
	resource.AddFile("materials/models/crt/bezel_mat.vmt")
	resource.AddFile("materials/models/crt/screen_mat.vmt")
	resource.AddFile("materials/models/crt/crt_diffuse.vtf")
	resource.AddFile("models/crt/crt.mdl")
	
	resource.AddFile("materials/models/holo/screen_mat.vmt")
	resource.AddFile("materials/models/holo/frame_mat.vmt")
	resource.AddFile("materials/models/holo/holo_diffuse.vtf")
	resource.AddFile("models/holo/holo.mdl")
	
	hook.Add("Initialize", "init_physgun_monitor", function()
		util.AddNetworkString("request_entity_mass")
		util.AddNetworkString("respond_entity_mass")
	end)

	--send the clients the mass of the objects they pick up
	--only the server can get access to a prop's physics object, and therefore its mass.
	hook.Add("PhysgunPickup", "physgun_get_mass", function(player, entity)
		if !IsValid(entity) or !IsValid(entity:GetPhysicsObject()) then return end
		net.Start("respond_entity_mass")
		net.WriteFloat(entity:GetPhysicsObject():GetMass())
		net.Send(player)
	end)
	
	--clients can also request the mass of a prop
	net.Receive("request_entity_mass", function(length, player)
		local targetEntity = net.ReadEntity()
		if !IsValid(targetEntity) or !IsValid(targetEntity:GetPhysicsObject()) then return end
		net.Start("respond_entity_mass")
		net.WriteFloat(targetEntity:GetPhysicsObject():GetMass())
		net.Send(player)
	end)
	
	--let the player know the server is running this addon too, so it's safe to ask for mass
	hook.Add("PlayerSpawn", "monitor_spawn_handler", function(player)
		net.Start("respond_entity_mass")
		net.WriteFloat(0)
		net.Send(player)
	end)
end

if CLIENT then
	--convars
	CreateClientConVar("physmon_flicker", "1", true, false, "should screen flicker?")
	CreateClientConVar("physmon_camview", "1", true, false, "show the camera view on screen?")
	CreateClientConVar("physmon_draw_wireframe", "1", true, false, "draw the wireframe model?")
	CreateClientConVar("physmon_wirespin_speed", "100", true, false, "how fast does the wire model spin?  (can be negative for reverse direction)")
	CreateClientConVar("physmon_wirescan_speed", "200", true, false, "how fast does the wire model \"scan\"?")
	CreateClientConVar("physmon_print_speed", "100", true, false, "how fast does the text print out?")
	CreateClientConVar("physmon_43_mode", "0", true, false, "move the monitor to the left to accomodate 4:3 displays?")
	CreateClientConVar("physmon_model_id", "1", true, false, "which model to use")

	local monitorViewmodel
	--selection info
	local targetEntity	--this is the entity to display info for
	local mass = -1	--the mass of the target entity.  this has to be retreived from the server
	local traceInfo
	--screen
	local SCREEN_WIDTH = 400	--display is 4:3 aspect ratio, and pretty small, so go with nice round numbers
	local SCREEN_HEIGHT = 300
	local BACKGROUND_COLOR = Color(20, 20, 20)	--color to clear the screen with
	local SCREEN_RENDER_TARGET = GetRenderTarget("physgun_monitor_screen", SCREEN_WIDTH, SCREEN_HEIGHT)	--create screen render target
	--fonts
	surface.CreateFont("physgun_monitor_title", {
		font	= "Courier New",
		size	= 60,
		weight	= 900
	})
	surface.CreateFont("physgun_monitor_regular", {
		font = "Courier New",
		size = 20,
		weight = 500
	})
	local fontColor
	local TITLE_FONT_SIZE = 60
	local NORMAL_FONT_SIZE = 30
	local TEXT_X_OFFSET = 0.025
	--artifact
	local scanArtifactPos = 0			--makes a lighter rectangle that moves down the screen like when you take a video of a CRT monitor
	local SCAN_ARTIFACT_LUMINANCE = 100	--how much lighter the artifact makes the screen (0-255)
	local SCAN_ARTIFACT_SIZE = 100		--how wide the artifact it (larger rectangle on y axis)
	local SCAN_ARTIFACT_SPEED = 150		--how fast the artifact moves down the screen
	--screen flicker
	local flickerElapsed = 0		--how long it has been since the last randomization
	local flickerValue = 0			--the randomized value to lighten/darken the screen by
	local FLICKER_AMOUNT = 1		--how intense the flicker is (0-255)
	local FLICKER_PERIOD = 0.017	--time before flicker value is re-randomized in seconds (0.017 is ~60hz)
	--wireframe
	local wireframeEntity	--this is the entity rendered for the wireframe background
	
	--timing
	local selectionElapsed = 0	--how much text has printed so far (also used for the wireframe scan effect for convenience.  it's a float)
	local prevDrawTime = 0	--store time since last draw to calculate deltatime for calculations
	
	local function CreateMonitor()
		--don't create multiple
		if IsValid(monitorViewmodel) then return end
		--don't create if the player isn't using a physgun
		if LocalPlayer():GetActiveWeapon():GetClass() != "weapon_physgun" then return end
		
		--create the monitor and attach it to the physgun's base bone so it animates properly
		if GetConVar("physmon_model_id"):GetInt() == 1 then
			monitorViewmodel = ClientsideModel(HOMEBREW_MODEL_PATH, RENDERGROUP_VIEWMODEL)
		elseif GetConVar("physmon_model_id"):GetInt() == 2 then
			monitorViewmodel = ClientsideModel(CRT_MODEL_PATH, RENDERGROUP_VIEWMODEL)
		elseif GetConVar("physmon_model_id"):GetInt() == 3 then
			monitorViewmodel = ClientsideModel(HOLO_MODEL_PATH, RENDERGROUP_VIEWMODEL)
		end
		--monitorViewmodel:FollowBone(LocalPlayer():GetViewModel(), LocalPlayer():GetViewModel():LookupBone("Base"))
		monitorViewmodel:SetLocalPos(Vector(0, 0, 0))
		monitorViewmodel:SetLocalAngles(Angle(0, 0, 0))
		
		--don't draw automatically, we have to wait and draw it manually in post viewmodel or it'll be a mess of clipping and incorrect FOV
		monitorViewmodel:SetNoDraw(true)
	end
	
	local function DestroyMonitor()
		if !IsValid(monitorViewmodel) then return end
		monitorViewmodel:Remove()
		heldEntity = nil
		targetEntity = nil
	end
	
	local function SetWireframeEntity(entity)
		if !IsValid(entity) then
			wireframeEntity = nil
			return
		end
	
		if !IsValid(wireframeEntity) then
			wireframeEntity = ClientsideModel(entity:GetModel(), RENDERGROUP_TRANSLUCENT)
			wireframeEntity:SetMaterial("models/wireframe")
			wireframeEntity:SetNoDraw(true)
		end
	
		--if the target is an effect, show the attached entity's model
		if entity:GetClass() == "prop_effect" then
			local effectEntity = entity.AttachedEntity
			if IsValid(effectEntity) then
				wireframeEntity:SetModel(effectEntity:GetModel())
			end
		else		
			wireframeEntity:SetModel(entity:GetModel())
		end
	end
	
	local activeWeapon
	local isAlive
	local isInVehicle
	local prevModelId
	local prevWeapon
	local prevAlive
	local prevInVehicle
	local heldEntity
	hook.Add("Think", "monitor_think_handler", function()
		--debug - make sure there is only one monitor at all times!
		--[[for key, value in ipairs(ents.FindByModel(CRT_MODEL_PATH)) do
			print(key, value)
		end]]
	
		activeWeapon = LocalPlayer():GetActiveWeapon()
		isAlive = LocalPlayer():Alive()
		isInVehicle = LocalPlayer():InVehicle()
		
		--check validity of weapons
		if !IsValid(activeWeapon) then return end
		if !IsValid(prevWeapon) then prevWeapon = activeWeapon end
		
		--remake the monitor when the player accepts a new model.
		local currentModelId = GetConVar("physmon_model_id"):GetInt()
		if prevModelId != currentModelId then
			DestroyMonitor()
			CreateMonitor()
			prevModelId = currentModelId
		end
		
		--create monitor when player spawns
		if isAlive and !prevAlive then
			CreateMonitor()
			--also the time when the player's physgun color could be changed, so recalculate font color now
			fontColor = LocalPlayer():GetWeaponColor()	--for some reason this returns a vector instead of a color struct
			local h, s, v = ColorToHSV(Color(fontColor.x * 255, fontColor.y * 255, fontColor.z * 255))
			v = 0.75--math.Clamp(v, 0.5, 0.75)	--make sure font color isn't too light or dark to read
			fontColor = HSVToColor(h, s, v)
		end
		--destroy screen when player dies
		if !isAlive and prevAlive then DestroyMonitor() end
		--create when leaving a vehicle
		if !isInVehicle and prevInVehicle then CreateMonitor() end
		--destroy when entering vehicle
		if isInVehicle and !prevInVehicle then DestroyMonitor() end
		--create/destroy when switching to/away from physgun
		if activeWeapon != prevWeapon then
			if activeWeapon:GetClass() == "weapon_physgun" then CreateMonitor() end
			if prevWeapon:GetClass() == "weapon_physgun" then DestroyMonitor() end
		end
		
		--check if we're picking something up.  needs to be checked every frame in case we are dragging the beam over something, just checking on press won't work.
		if activeWeapon:GetClass() == "weapon_physgun" and LocalPlayer():KeyDown(IN_ATTACK) then 
			if !heldEntity then
				local eyeEntity = LocalPlayer():GetEyeTrace().Entity
				
				--make sure this is something we can pick up
				if IsValid(eyeEntity) and !string.find(eyeEntity:GetClass(), "func_") and !string.find(eyeEntity:GetClass(), "BaseEntity") then
					heldEntity = eyeEntity

					if heldEntity != targetEntity then
						targetEntity = heldEntity
						SetWireframeEntity(targetEntity)

						selectionElapsed = 0
					end
				end
			end
		else
			heldEntity = nil
		end
		
		prevWeapon = activeWeapon
		prevAlive = isAlive
		prevInVehicle = isInVehicle
	end)
	
	hook.Add("KeyPress", "monitor_keypress_handler", function(player, key)
		if !IsValid(player) or player != LocalPlayer() then return end
		--this gets called multiple times in multiplayer, so make sure this is the first
		if !game.SinglePlayer() and !IsFirstTimePredicted() then return end
		--nothing to do if we aren't holding the physgun
		if !IsValid(activeWeapon) or !activeWeapon:GetClass() == "weapon_physgun" then return end
		
		--reload clears the entity if there is one, and if there isn't it selects the local player
		if key == IN_RELOAD then
			if IsValid(targetEntity) or (targetEntity and targetEntity:IsWorld()) then
				targetEntity = nil
				SetWireframeEntity(nil)
			else
				targetEntity = LocalPlayer()
				SetWireframeEntity(nil)
				
				--we need to ask the server for the mass because there's no way to get it clientside as far as I know.  this obviously won't work in multiplayer
				if mass != -1 then
					net.Start("request_entity_mass")
					net.WriteEntity(targetEntity)
					net.SendToServer()
				end
			
				--reset print progress so text prints out again
				selectionElapsed = 0
			end
		end

		--select whatever is under the crosshair
		if key == IN_ATTACK2 and !heldEntity then
			traceInfo = LocalPlayer():GetEyeTrace()
			local eyeEntity = traceInfo.Entity
			if !eyeEntity or eyeEntity == NULL then return end
			if string.find(eyeEntity:GetClass(), "func_") or string.find(eyeEntity:GetClass(), "BaseEntity") then
				if targetEntity != game.GetWorld() then
					targetEntity = game.GetWorld()
					selectionElapsed = 0
				end
			elseif eyeEntity and eyeEntity != targetEntity then
				targetEntity = eyeEntity
				SetWireframeEntity(targetEntity)

				if mass != -1 then
					net.Start("request_entity_mass")
					net.WriteEntity(targetEntity)
					net.SendToServer()
				end
				
				selectionElapsed = 0
			end
		end
	end)
	
	--get whether or not to draw the player.  we need this because normally the player doesn't appear on the camera view because technically it's still firstperson at that point
	local drawingCameraScreen = false
	if GetConVar("physmon_camview"):GetBool() then
		hook.Add("ShouldDrawLocalPlayer", "monitor_draw_local_player", function(player)
			--print(drawingCameraScreen)
			return drawingCameraScreen
		end)
	end
	
	local function DrawScrollingText(text, yPos)
		local width, height = surface.GetTextSize(text)
		width = width + 64	--space text out

		yPos = yPos - height / 2 --center text to y position

		local xPos = RealTime() * 250 % width * -1

		while (xPos < SCREEN_WIDTH) do
			surface.SetTextPos(xPos, yPos)
			surface.DrawText(text)

			xPos = xPos + width
		end
	end
	
	--draws justified text on the screen
	local LEFT_JUSTIFIED, CENTER_JUSTIFIED, RIGHT_JUSTIFIED = 1, 2, 3	--pseudo-enum I guess
	local function DrawTextJustified(text, xPos, yPos, justification)
		local width = surface.GetTextSize(text)
		if justification == LEFT_JUSTIFIED then surface.SetTextPos(xPos, yPos)
		elseif justification == CENTER_JUSTIFIED then surface.SetTextPos(xPos - width / 2, yPos)
		elseif justification == RIGHT_JUSTIFIED then surface.SetTextPos(xPos - width, yPos) end
		surface.DrawText(text)
	end
	
	--if the target entity is a camera, we draw its view to the screen
	local function DrawCamView()
		--don't draw the model or halo since we will see that on screen
		local isNoDraw = targetEntity:GetNoDraw()
		local drawHalo = GetConVar("physgun_halo"):GetBool()
		targetEntity:SetNoDraw(true)
		RunConsoleCommand("physgun_halo", 0)
		
		--for some reason we have to call Start2D before drawing or ShouldDrawLocalPlayer will not even be called.
		--we can just end it immediately, but it has to be called.  It's times like these that I question everything I thought I knew about programming.
		cam.Start2D()
		cam.End2D()
		
		--draw the player's thirdperson
		drawingCameraScreen = true
		--this is really lagtastic on some maps.  maybe I should add an option for it?
		render.RenderView({
			origin = targetEntity:GetPos(),
			angles = targetEntity:GetAngles(),
			x = 0,
			y = 0,
			w = SCREEN_WIDTH,
			h = SCREEN_HEIGHT,
			drawhud = false,
			dopostprocess = false,
			drawviewmodel = false
		})
		drawingCameraScreen = false
		
		targetEntity:SetNoDraw(isNoDraw)
		RunConsoleCommand("physgun_halo", drawHalo)
	end
	
	--draws a spinning, isometric wireframe of the target entity on the background
	local function DrawWireframe(backgroundColor)
		--size of ortho camera should be a bit larger than the model's radius so it fits on screen with a little room to spare
		local orthoSize = targetEntity:GetModelRadius() + 10
		cam.Start3D(Vector(0, 0, 0), Angle(0, 0, 0), 90, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT)	--position camera at map origin
		cam.StartOrthoView(-orthoSize, orthoSize, orthoSize, -orthoSize)

		local drawMatrix = Matrix()
		--many models have their origins at the bottom or somewhere else that causes them to render at an offset.  we can correct that most of the time by subtracting the center of their bounding box
		drawMatrix:Translate(wireframeEntity:GetPos() - wireframeEntity:OBBCenter())
		--pitch forward for an isometric view
		drawMatrix:Rotate(Angle(-30, 0, 0))
		--spin about the up axis
		drawMatrix:Rotate(Angle(0, CurTime() * GetConVar("physmon_wirespin_speed"):GetFloat(), 0))

		wireframeEntity:EnableMatrix("RenderMultiply", drawMatrix)
		wireframeEntity:DrawModel()
		wireframeEntity:DisableMatrix("RenderMultiply")

		cam.EndOrthoView()
		cam.End3D()
		
		cam.Start2D()
		
		--wireframe "scan" effect
		if GetConVar("physmon_wirescan_speed"):GetInt() != 0 then
			surface.SetDrawColor(backgroundColor)
			surface.DrawRect(0, selectionElapsed * GetConVar("physmon_wirescan_speed"):GetFloat(), SCREEN_WIDTH, SCREEN_HEIGHT)
			surface.SetDrawColor(255, 255, 255, 255)
			surface.DrawLine(0, selectionElapsed * GetConVar("physmon_wirescan_speed"):GetFloat(), SCREEN_WIDTH, selectionElapsed * GetConVar("physmon_wirescan_speed"):GetFloat())
		end
		
		--darken wireframe
		surface.SetDrawColor(backgroundColor.r, backgroundColor.g, backgroundColor.b, 250)
		surface.DrawRect(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT)
		
		cam.End2D()
	end
	
	--print out useful information about the target entity
	local function DrawEntityPrintout()
		cam.Start2D()

		surface.SetTextColor(fontColor)

		--title box
		surface.SetDrawColor(Color(255, 255, 255, 5))
		surface.DrawRect(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT * 0.2)
	
		--name
		local textCache
		if targetEntity:IsPlayer() then
			textCache = targetEntity:Nick()
		elseif targetEntity:GetClass() == "prop_effect" then
			local effectEntity = targetEntity.AttachedEntity
			if IsValid(effectEntity) then
				textCache = string.sub(effectEntity:GetModel(), string.find(effectEntity:GetModel(), "/[^/]*$") + 1, -5)
			end
		else
			--get name from model by substring everything from the last slash to five before the end.  omits the path and extension
			textCache = string.sub(targetEntity:GetModel(), string.find(targetEntity:GetModel(), "/[^/]*$") + 1, -5)
		end
		
		surface.SetFont("physgun_monitor_title")
		DrawScrollingText(textCache, 30, SCREEN_WIDTH)
	
		surface.SetFont("physgun_monitor_regular")
		local printLimit = selectionElapsed * GetConVar("physmon_print_speed"):GetInt()
		if printLimit == 0 then printLimit = 1000 end
		
		--id
		textCache = "id: " .. targetEntity:EntIndex()
		--selectionElapsed increases over time, so the string will print more and more to make it look like and old computer slowly printing
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.3, LEFT_JUSTIFIED)
	
		--class
		textCache = "class: " .. targetEntity:GetClass()
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.4, LEFT_JUSTIFIED)
	
		--position
		--max map size is -16384 to 16384, so we only need five digits for each coordinate
		--with three places of precision and the decimal point, that's 9 characters long, so format is %9.3f
		textCache = "pos: " .. string.format("%9.3f", targetEntity:GetPos().x) .. " " .. string.format("%9.3f", targetEntity:GetPos().y) .. " " .. string.format("%9.3f", targetEntity:GetPos().z)
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.5, LEFT_JUSTIFIED)
	
		--angles
		textCache = "ang: " .. string.format("%9.3f", targetEntity:GetAngles().pitch) .. " " .. string.format("%9.3f", targetEntity:GetAngles().yaw) .. " " .. string.format("%9.3f", targetEntity:GetAngles().roll)
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.6, LEFT_JUSTIFIED)
	
		--velocity
		textCache = "vel: " .. string.format("%9.3f", targetEntity:GetVelocity().x) .. " " .. string.format("%9.3f", targetEntity:GetVelocity().y) .. " " .. string.format("%9.3f", targetEntity:GetVelocity().z)
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.7, LEFT_JUSTIFIED)
	
		--mass
		if mass == -1 then
			textCache = "mass: --unavailable on client--"
		else
			textCache = "mass: " .. string.format("%9.3f", mass)
		end
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.8, LEFT_JUSTIFIED)
	
		cam.End2D()
	end
	
	--same as above, but different info for the world instead of normal entities
	local function DrawWorldPrintout()
		cam.Start2D()
		
		surface.SetTextColor(fontColor)
		
		--title box
		surface.SetDrawColor(Color(255, 255, 255, 5))
		surface.DrawRect(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT * 0.2)
	
		--name
		local textCache = game.GetMap()
		surface.SetFont("physgun_monitor_title")
		DrawScrollingText(textCache, 30, SCREEN_WIDTH)
		
		surface.SetFont("physgun_monitor_regular")
		local printLimit = selectionElapsed * GetConVar("physmon_print_speed"):GetInt()
		if printLimit == 0 then printLimit = 1000 end
		
		--brush's material
		local lastSlashInd = string.find(traceInfo.HitTexture, "/[^/]*$")
		if lastSlashInd then lastSlashInd = lastSlashInd + 1 else lastSlashInd = 0 end
		textCache = "texture: " .. string.sub(traceInfo.HitTexture, lastSlashInd, -1)
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.3, LEFT_JUSTIFIED)
		
		--texture path
		textCache = "texpath: " .. string.sub(traceInfo.HitTexture, 0, lastSlashInd - 1)
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.4, LEFT_JUSTIFIED)
		
		--hit position
		textCache = "hitpos: " .. string.format("%5.3f", traceInfo.HitPos.x) .. " " .. string.format("%5.3f", traceInfo.HitPos.y) .. " " .. string.format("%5.3f", traceInfo.HitPos.z)
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.5, LEFT_JUSTIFIED)
		
		--surface normal
		textCache = "normal: " .. string.format("%5.3f", traceInfo.HitNormal.x) .. " " .. string.format("%5.3f", traceInfo.HitNormal.y) .. " " .. string.format("%5.3f", traceInfo.HitNormal.z)
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.6, LEFT_JUSTIFIED)
		
		--distance from the player's eye to the point under crosshair
		textCache = "distance: " .. string.format("%9.3f", (traceInfo.HitPos - traceInfo.StartPos):Length())
		textCache = string.sub(textCache, 0, printLimit)
		DrawTextJustified(textCache, SCREEN_WIDTH * TEXT_X_OFFSET, SCREEN_HEIGHT * 0.7, LEFT_JUSTIFIED)
		
		cam.End2D()
	end

	--draw the screen to the rendertexture
	local homebrewScreenMaterial = Material("models/homebrew/screen_mat")
	local crtScreenMaterial = Material("models/crt/screen_mat")
	local holoScreenMaterial = Material("models/holo/screen_mat")
	local function RenderScreen()
		local deltaTime = CurTime() - prevDrawTime
		selectionElapsed = selectionElapsed + deltaTime

		--choose which screen material to use
		local screenMaterial
		if GetConVar("physmon_model_id"):GetInt() == 1 then
			screenMaterial = homebrewScreenMaterial
		elseif GetConVar("physmon_model_id"):GetInt() == 2 then
			screenMaterial = crtScreenMaterial
		elseif GetConVar("physmon_model_id"):GetInt() == 3 then
			screenMaterial = holoScreenMaterial
		end
	
		--draw to the screen
		render.PushRenderTarget(SCREEN_RENDER_TARGET, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT)
		screenMaterial:SetTexture("$basetexture", SCREEN_RENDER_TARGET)

		--clear buffers
		render.Clear(0, 0, 0, 0)
		render.ClearDepth()
		render.ClearStencil()

		--flicker
		local backgroundWithFlicker
		if GetConVar("physmon_flicker"):GetBool() then
			flickerElapsed = flickerElapsed + deltaTime
			if flickerElapsed > FLICKER_PERIOD then
				flickerValue = math.random(0, 1)
				flickerElapsed = 0
			end
			backgroundWithFlicker = Color(BACKGROUND_COLOR.r + flickerValue, BACKGROUND_COLOR.g + flickerValue, BACKGROUND_COLOR.b + flickerValue, 255)
		else
			backgroundWithFlicker = BACKGROUND_COLOR
		end
		
		surface.SetDrawColor(backgroundWithFlicker)
		cam.Start2D()
		surface.DrawRect(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT)
		cam.End2D()
		
		--draw info
		if IsValid(targetEntity) then
			if targetEntity:GetClass() == "gmod_cameraprop" and GetConVar("physmon_camview"):GetBool() then
				DrawCamView()
			else
				if GetConVar("physmon_draw_wireframe"):GetBool() and IsValid(wireframeEntity) then
					DrawWireframe(backgroundWithFlicker)
				end
				
				DrawEntityPrintout()
			end
		elseif targetEntity and targetEntity:IsWorld() then
			DrawWorldPrintout()
		end
		
		--draw scan artifact
		if GetConVar("physmon_flicker"):GetBool() then
			cam.Start2D()
			surface.SetDrawColor(SCAN_ARTIFACT_LUMINANCE, SCAN_ARTIFACT_LUMINANCE, SCAN_ARTIFACT_LUMINANCE, 1 + flickerValue)
			surface.DrawRect(0, scanArtifactPos, SCREEN_WIDTH, SCAN_ARTIFACT_SIZE)
			scanArtifactPos = scanArtifactPos + deltaTime * SCAN_ARTIFACT_SPEED
			if scanArtifactPos > SCREEN_HEIGHT then scanArtifactPos = -SCAN_ARTIFACT_SIZE end
			cam.End2D()
		end
		
		render.PopRenderTarget()
		
		prevDrawTime = CurTime()
	end
	
	local boneID
	hook.Add("PostDrawViewModel", "draw_view_monitor", function(weaponviewmodel, player, weapon)
		if !IsValid(monitorViewmodel) then return end
		if !boneID then boneID = LocalPlayer():GetViewModel():LookupBone("Base") end
		if !boneID then return end
		
		--we can't parent to the bone because this causes the model to jump around when the ping spikes on servers for some reason.
		--so instead, just copy the bone's transform matrix immediately before rendering each frame
		local viewMatrix = LocalPlayer():GetViewModel():GetBoneMatrix(boneID)
		--some custom sweps will cause this to be called for one frame after switching to them?  quit here in case that happens (need to find out why though)
		if !viewMatrix then return end
		
		if GetConVar("physmon_43_mode"):GetBool() then
			viewMatrix:Translate(Vector(3, 2, -10))
		else
			viewMatrix:Translate(Vector(-1, 2, -10))
		end
		viewMatrix:Rotate(Angle(90 - 25, 90, 0))
		
		RenderScreen()
		
		monitorViewmodel:EnableMatrix("RenderMultiply", viewMatrix)
		cam.IgnoreZ(true)
		monitorViewmodel:DrawModel()
		monitorViewmodel:DisableMatrix("RenderMultiply")
	end)
	
	--server will send us the mass of the picked up or right clicked entity
	net.Receive("respond_entity_mass", function()
		mass = net.ReadFloat()
	end)
	
	--settings panel accessible through the context menu
	local SETTINGS_FRAME_HEIGHT = 415
	local SETTINGS_FRAME_WIDTH = 405
	local isWindowOpen = false
	local function CreateInfoMonitorSettingsWindow(player, command, args, argStr)
		isWindowOpen = true
	
		local frame = vgui.Create("DFrame")
		frame:SetTitle("Physgun Info Monitor Settings")
		frame:SetSize(SETTINGS_FRAME_WIDTH, SETTINGS_FRAME_HEIGHT)
		frame:Center()
		frame:MakePopup()
		frame.OnClose = function()
			isWindowOpen = false
		end
		
		--camview
		local camviewPanel = vgui.Create("DPanel", frame)
		camviewPanel:DockMargin(0, 5, 0, 0)
		camviewPanel:DockPadding(5, 5, 5, 5)
		camviewPanel:Dock(TOP)
		local camviewLabel = vgui.Create("DLabel", camviewPanel)
		camviewLabel:SetDark(true)
		camviewLabel:SetText("Show camera views")
		camviewLabel:SizeToContents()
		camviewLabel:Dock(LEFT)
		local camviewCheckbox = vgui.Create("DCheckBox", camviewPanel)
		camviewCheckbox:Dock(RIGHT)
		camviewCheckbox:SetChecked(GetConVar("physmon_camview"):GetBool())
		
		--monitor flicker
		local flickerPanel = vgui.Create("DPanel", frame)
		flickerPanel:DockMargin(0, 5, 0, 0)
		flickerPanel:DockPadding(5, 5, 5, 5)
		flickerPanel:Dock(TOP)
		local flickerLabel = vgui.Create("DLabel", flickerPanel)
		flickerLabel:SetDark(true)
		flickerLabel:SetText("Monitor flicker")
		flickerLabel:SizeToContents()
		flickerLabel:Dock(LEFT)
		local flickerCheckbox = vgui.Create("DCheckBox", flickerPanel)
		flickerCheckbox:Dock(RIGHT)
		flickerCheckbox:SetChecked(GetConVar("physmon_flicker"):GetBool())
		
		--draw wireframe
		local wireframePanel = vgui.Create("DPanel", frame)
		wireframePanel:DockMargin(0, 5, 0, 0)
		wireframePanel:DockPadding(5, 5, 5, 5)
		wireframePanel:Dock(TOP)
		local wireframeLabel = vgui.Create("DLabel", wireframePanel)
		wireframeLabel:SetDark(true)
		wireframeLabel:SetText("Draw wireframe model")
		wireframeLabel:SizeToContents()
		wireframeLabel:Dock(LEFT)
		local wireframeCheckbox = vgui.Create("DCheckBox", wireframePanel)
		wireframeCheckbox:Dock(RIGHT)
		wireframeCheckbox:SetChecked(GetConVar("physmon_draw_wireframe"):GetBool())
		
		--4:3 mode
		local _43Panel = vgui.Create("DPanel", frame)
		_43Panel:DockMargin(0, 5, 0, 0)
		_43Panel:DockPadding(5, 5, 5, 5)
		_43Panel:Dock(TOP)
		local _43Label = vgui.Create("DLabel", _43Panel)
		_43Label:SetDark(true)
		_43Label:SetText("4:3 mode for non-widescreen monitors")
		_43Label:SizeToContents()
		_43Label:Dock(LEFT)
		local _43Checkbox = vgui.Create("DCheckBox", _43Panel)
		_43Checkbox:Dock(RIGHT)
		_43Checkbox:SetChecked(GetConVar("physmon_43_mode"):GetBool())
		
		--wirespin speed
		local spinSpeedPanel = vgui.Create("DPanel", frame)
		spinSpeedPanel:DockMargin(0, 5, 0, 0)
		spinSpeedPanel:DockPadding(5, 5, 5, 5)
		spinSpeedPanel:Dock(TOP)
		local spinSpeedlabel = vgui.Create("DLabel", spinSpeedPanel)
		spinSpeedlabel:SetDark(true)
		spinSpeedlabel:SetText("Wireframe spin speed")
		spinSpeedlabel:SizeToContents()
		spinSpeedlabel:Dock(LEFT)
		local spinSpeedSlider = vgui.Create("DNumSlider", spinSpeedPanel)
		spinSpeedSlider:Dock(FILL)
		spinSpeedSlider:SetMin(-255)
		spinSpeedSlider:SetMax(255)
		spinSpeedSlider:SetDecimals(0)
		spinSpeedSlider:SetValue(GetConVar("physmon_wirespin_speed"):GetInt())
		
		--wirescan speed
		local scanSpeedPanel = vgui.Create("DPanel", frame)
		scanSpeedPanel:DockMargin(0, 5, 0, 0)
		scanSpeedPanel:DockPadding(5, 5, 5, 5)
		scanSpeedPanel:Dock(TOP)
		local scanSpeedlabel = vgui.Create("DLabel", scanSpeedPanel)
		scanSpeedlabel:SetDark(true)
		scanSpeedlabel:SetText("Wireframe scan speed")
		scanSpeedlabel:SizeToContents()
		scanSpeedlabel:Dock(LEFT)
		local scanSpeedSlider = vgui.Create("DNumSlider", scanSpeedPanel)
		scanSpeedSlider:Dock(FILL)
		scanSpeedSlider:SetMin(0)
		scanSpeedSlider:SetMax(1000)
		scanSpeedSlider:SetDecimals(0)
		scanSpeedSlider:SetValue(GetConVar("physmon_wirescan_speed"):GetInt())
		
		--print speed
		local printSpeedPanel = vgui.Create("DPanel", frame)
		printSpeedPanel:DockMargin(0, 5, 0, 0)
		printSpeedPanel:DockPadding(5, 5, 5, 5)
		printSpeedPanel:Dock(TOP)
		local printSpeedlabel = vgui.Create("DLabel", printSpeedPanel)
		printSpeedlabel:SetDark(true)
		printSpeedlabel:SetText("Print speed")
		printSpeedlabel:SizeToContents()
		printSpeedlabel:Dock(LEFT)
		local printSpeedSlider = vgui.Create("DNumSlider", printSpeedPanel)
		printSpeedSlider:Dock(FILL)
		printSpeedSlider:SetMin(0)
		printSpeedSlider:SetMax(255)
		printSpeedSlider:SetDecimals(0)
		printSpeedSlider:SetValue(GetConVar("physmon_print_speed"):GetInt())
		
		--model
		local modelPanel = vgui.Create("DPanel", frame)
		modelPanel:DockMargin(0, 5, 0, 0)
		modelPanel:DockPadding(5, 5, 5, 5)
		modelPanel:Dock(FILL)
		
		local modelList = vgui.Create("DListView", modelPanel)
		modelList:Dock(FILL)
		modelList:SetMultiSelect(false)
		modelList:AddColumn("Model")
		modelList:AddLine("Homebrew")
		modelList:AddLine("Tube")
		modelList:AddLine("Simple")
		modelList:SelectItem(modelList:GetLine(GetConVar("physmon_model_id"):GetInt()))
		
		--close buttons
		local buttonPanel = vgui.Create("DPanel", frame)
		buttonPanel:Dock(BOTTOM)
		buttonPanel:SetBackgroundColor(Color(0, 0, 0, 0))
		local cancelButton = vgui.Create("DButton", buttonPanel)
		cancelButton:Dock(LEFT)
		cancelButton:SetText("Cancel")
		cancelButton.DoClick = function()
			frame:Close()
		end
		local acceptButton = vgui.Create("DButton", buttonPanel)
		acceptButton:Dock(FILL)
		acceptButton:DockMargin(5, 0, 0, 0)
		acceptButton:SetText("Accept")
		acceptButton.DoClick = function()
			RunConsoleCommand("physmon_flicker", flickerCheckbox:GetChecked() and 1 or 0)
			RunConsoleCommand("physmon_camview", camviewCheckbox:GetChecked() and 1 or 0)
			RunConsoleCommand("physmon_draw_wireframe", wireframeCheckbox:GetChecked() and 1 or 0)
			RunConsoleCommand("physmon_43_mode", _43Checkbox:GetChecked() and 1 or 0)
			RunConsoleCommand("physmon_print_speed", printSpeedSlider:GetValue())
			RunConsoleCommand("physmon_wirespin_speed", spinSpeedSlider:GetValue())
			RunConsoleCommand("physmon_wirescan_speed", scanSpeedSlider:GetValue())
			RunConsoleCommand("physmon_model_id", modelList:GetSelectedLine())

			--make this hook won't do anything unless we want can views, and it causes problem with other addons.
			--turn it off when we aren't using it so players can use this addon without conflicts
			if camviewCheckbox:GetChecked() then
				hook.Add("ShouldDrawLocalPlayer", "monitor_draw_local_player", function(player)
					return drawingCameraScreen
				end)
			else
				hook.Remove("ShouldDrawLocalPlayer", "monitor_draw_local_player")
			end

			frame:Close()
		end
	end
	
	--add the settings window to the context desktop
	list.Set("DesktopWindows", "pim_editor", {
		title		= "Info Monitor",
		icon		= "materials/physgun_info_monitor/context_icon.png",
		width		= 1,
		height		= 1,
		onewindow	= false,
		init		= function(icon, window)
			window:Remove()
			if !isWindowOpen then
				CreateInfoMonitorSettingsWindow()
			end
		end})
end
